\ This software is free for use and modification by anyone for any purpose
\ with no restrictions or source identification of any kind.

\ Nov 6 2016 Douglas B. Hoffman
\ dhoffman888@gmail.com


[undefined] floats [if]
cr .( Warning: Floating Point is not present, some library code will not be loaded)
[then]

\ utility words
: isdigit ( char -- flag ) \ true if 0 thru 9, false otherwise
  48 58 within ;
: isalpha ( char -- flag ) \ true if a-z or A-Z, false otherwise
  dup  65  91 within
  swap 97 123 within or ;

\ convert given stringto a signed decimal integer
: >integer ( addr len  -- n t | f)
  0 1 0 locals| neg? dec accum len addr |
  addr c@ [char] - = to neg?
  len 1+ neg? + 1 ?do
   addr len + i - c@ dup isdigit
     if ( it's a digit 0 thru 9 )
        48 - dec * accum + to accum
        dec 10 * to dec
     else [char] . <> if false unloop exit then
     then
  loop accum neg? if negate then true ;


:class var  
  ivar data 
 :m init: 0 data ! ;m
 :m !: ( n -- )  data ! ;m
 :m @: ( -- n )  data @ ;m
 :m +: ( n -- )  data +! ;m
 :m p: ( -- ) self @: . ;m \ print self
 :m =: ( obj -- flag ) @: self @: = ;m
 \ a copy: object will always be instantiated in the heap
 :m copy: ( -- obj ) self @: heap> var dup >r !: r> ;m
;class

[defined] floats [if]

:class fvar    
  1 floats bytes data
 :m init: 0e data f! ;m
 :m !: ( r -- )  data f! ;m
 :m @: ( -- r )  data f@ ;m
 :m +: ( r -- )  data f@ f+ data f! ;m
 :m p: ( -- ) self @: f. ;m \ print self
 :m =: ( obj -- flag ) @: self @: f= ;m
 :m copy: ( -- obj ) self @: heap> fvar dup >r !: r> ;m
;class
 

[then]

:class bool    
  1 bytes data
  :m set: -1  data c! ;m
  :m clear: 0  data c! ;m
  :m @: ( -- flag )  data c@ 255 = ;m
  :m p: self @: . ;m
;class

:class ptr 
 ivar data
 ivar size \ size, in bytes, of memory allocated\allotted
 :m size: ( -- n )  size @ ;m

 :m ptr: ( -- ptr ) data @ ;m

 :m free:
    data @ ?dup if free' throw' 0 data ! then  0  size ! ;m

 :m new:  ( size -- ) 
    self free:  
    dup allocate' throw'  data !   size ! ;m

 :m resize:  ( newsize -- )
    dup >r
    data @ size @ newsize' throw'  data !   r> size !  ;m
    
 :m init: 0 data ! 0  size ! ;m
 
 \ (add) (@) (p) (+) all define code that is common to the subclasses string and dicString
 : (add) ( addr len -- )  \ add text to end of string
    dup ( addr-src len len )
    size @ dup >r + self resize:  \ addr-src len
    data @ r> + ( addr-src len dest) swap cmove ;
 : (@) ( -- addr len ) data @   size @ ;
 : (p) self @: type ;
 : (+) ( char -- ) \ add char to end of string
   size @ 1+ self resize:   self @: + 1- c! ;
;class



:class dicPtr <super ptr
\ ivar data in the superclass will contain 0 or the allotted pointer
 ivar maxsize
 :m free: ;m \ does nothing
 :m new: ;m \ does nothing
 : !max ( n -- ) dup maxsize !
     here data ! allot ;

 : size-check ( n -- n)
        dup maxsize @ > abort" dicPtr size exceeded" ;
 :m resize: ( newsize -- )
     size-check size ! ;m
;class


:class string <super ptr
 :m !: ( addr len -- )
   dup super new: ( addr len ) data @ swap ( addr self len ) cmove ;m
 :m add: ( addr len -- ) \ add text to end of string
    (add) ;m
 :m @: ( -- addr len ) (@) ;m
 :m p: (p) ;m
 :m +: ( char -- ) \ add char to end of string
      (+) ;m
;class

\ don't use heap> or heap: with the following class
:class dicString <super dicPtr
 :m init: ( addr len max-size -- ) super init: !max self !: ;m
 :m !: ( addr len -- ) dup super resize: data @  swap cmove ;m
 :m add: ( addr len -- ) \ add text to end of string
    (add) ;m
 :m @: ( -- addr len ) (@) ;m
 :m p: (p) ;m
 :m +: ( char -- ) \ add char to end of string
      (+) ;m
;class


0 [if]


s" hello" 20 dicString d     

ok

d p: 
hello
ok

s" world" d add: d p: 
helloworld
ok

s" woooowooooowoooooo" d add: 

d p: helloworldok
s" bye" d !: d p: byeok

[then]


\ dynamically resizable arrays
:class 1-array <super ptr
 ivar #elems
 ivar current-idx
 ivar elemSize
 ivar sort-xt \  for sort comparison
 ivar map-xt
 ivar temp-array
 ivar =xt

fmsCheck? [if]
  :m ?idx: ( idx -- idx )
    dup 0< if dup #elems @ + 0= if  exit then then
    dup abs #elems @ < 0= abort" index out of range" ;m 
    
  : ^elem ( idx -- addr ) self ?idx: dup 0< if #elems @ + then  elemSize @ * data @  + ; 
 [else]
  : ^elem ( idx -- addr )  dup 0< if #elems @ + then elemSize @ * data @  + ;
 [then]

 :m @elem: ( addr -- elem ) @ ;m
 :m uneach: 0  current-idx ! ;m
 :m each: ( -- elem true | false)
     current-idx @ dup  #elems @ <
    if 1  current-idx +!  ^elem  self @elem: true else drop false self uneach: then ;m
 :m init:  super init: self uneach: 1 cells  elemSize !
    0 #elems !
     ['] < sort-xt !
     0 self new:
     ['] = =xt !
     ;m
 :m new: ( #elems -- ) dup  #elems !  elemSize @ * super new: ;m
 :m size: ( -- n)  #elems @ ;m
 :m to: ( elem idx -- )  ^elem  ! ;m
 :m at: ( idx -- elem )  ^elem  @ ;m
 
 : (copy):
    #elems @ temp-array @ new:
    self size: 0 ?do i self at: i temp-array @ to: loop
    temp-array @ ;
 :m copy: ( -- heap-1-array )
    heap> 1-array temp-array !
    (copy): ;m

 :m .elem: ( elem -- ) . ;m
 :m p:  #elems @ 0 ?do cr i dup . self at:  self .elem: loop ;m
 :m resize: ( #elems -- ) dup  #elems !  elemSize @ * super resize: ;m
 :m delete: ( idx -- ) \ delete elem at idx and shrink array by 1 cell
    dup 1+  #elems @ <>
    IF dup >r 1+  ^elem ( src)
           r@  ^elem ( dest)  #elems @ r> -  elemSize @ * ( len) move
    ELSE drop
    THEN #elems @ 1 - self resize:
    ;m
 :m ^elem: ( idx -- addr ) ^elem ;m
 :m fill: ( n -- ) self size: 0 ?do dup i self to: loop drop ;m
 :m indexOf: ( n -- idx true | false )
    self size: 0 ?do dup i self at: =
          if drop i true unloop exit then loop drop false ;m
 :m add: ( n -- ) \ increase size of array by one and place n in last position
    #elems @ dup 1+ self resize:
    self to: ;m 

 :m !sort: ( xt -- ) sort-xt ! ;m
 
 :m insert: ( n idx -- ) \ increase size of array by one, place n at given index position
                         \ after having moved all ensuing elements up by one position.
    >r
    #elems @ 1+ self resize:
    r@ self ^elem: ( src) r@ 1+ self ^elem: ( dest) self size: r@ - elemSize @ * ( len) move
    r> self to: ;m

\ quicksort based on Wil Baden's code, for sorting integers

: lessthan ( a@ b@ -- ? )   sort-xt @ execute ;

: mid ( l r -- mid ) over - 2/ -cell and + ;

: exch ( addr1 addr2 -- ) dup @ >r over @ swap ! r> swap ! ;

: part ( l r -- l r r2 l2 )
  2dup mid @ >r ( r: pivot )
  2dup begin
    swap begin dup @  r@ lessthan while cell+ repeat
    swap begin r@ over @ lessthan while cell- repeat
    2dup <= if 2dup exch >r cell+ r> cell- then
  2dup > until  r> drop ;

: qsort ( l r -- )
  part  swap rot
  2dup < if recurse else 2drop then
  2dup < if recurse else 2drop then ;

 :m sort: \ performs a quicksort using the sort-xt for comparisons
    0 self ^elem: #elems @ ( array len -- )
    dup 2 < if 2drop exit then
    1- cells over + qsort ;m

 :m set: ( N0 N1 ... Nn n -- ) \ will first empty the array
    0 self new:
    0 ?do self add: loop ;m 
 :m last: ( -- n ) -1 self at: ;m
 :m each2: ( -- elem1 elem2 true | false)
     current-idx @ #elems @ 1- <
    if current-idx @ ^elem self @elem:
       current-idx @ 1+ ^elem self @elem:
       1  current-idx +! true
    else false self uneach: then ;m

  :m map: ( xt -- <varies> ) \ apply xt to each object in the array
                             \ items in array do not change
     super size: 0= IF drop exit THEN   
       >r
      BEGIN self each: WHILE r@ execute REPEAT r> drop ;m

  :m map-to: ( xt -- array ) \ apply xt to each object in the list
                           \ items in list are changed
     map-xt !
     self size: 0 ?do
      i self at: map-xt @ execute i self to: loop self ;m

 :m sum: ( -- n )
    0
    begin
      self each:
    while
      +
    repeat ;m
 :m average: ( -- n )
      self sum:
      #elems @ / ;m
 :m cmp: ( xt -- n )
    >r
    self each: drop
    begin
     self each:
    while
    2dup
     r@ execute if swap then drop
    repeat r> drop ;m
 :m max: ( -- max )
    self each: drop
    begin
     self each:
    while
    2dup
     < if swap then drop
    repeat ;m
 :m min: ( -- min )
    self each: drop
    begin
     self each:
    while
    2dup
     > if swap then drop
    repeat ;m

 :m +: ( array -- ) \ concatenate given array to the end of this array
    >r begin
         r@ each:
       while 
         self add:
       repeat r> drop ;m

:m reverse:
   self size: 2/ 0
   ?do i self at: i 1+ negate self at:
       i self to: i 1+ negate self to: loop ;m

:m keep-duplicates: ( -- list ) \ returns a new sorted list instantiated in the heap
   self sort:
   self
   heap> 1-array 
   over each: drop 0 locals| added last list2 |
   begin
     dup each:
   while
     dup last =xt @ execute if added 0= if list2 add: -1 to added else drop then
               else to last 0 to added
               then
   repeat drop list2 ;m 

:m remove-duplicates: ( -- list ) \ returns a new sorted list instantiated in the heap
   self sort:
   self
   heap> 1-array 
   over each: drop locals| last list2 |
   last list2 add:
   begin
     dup each:
   while
     dup last =xt @ execute if drop
                else dup list2 add: to last  
                then
   repeat drop list2 ;m 

;class


:class set <super 1-array

 :m add: ( n -- )
    dup self indexOf: if 2drop exit then
    super add: ;m

  :m to: ( elem idx -- )
     >r
     dup self indexOf: if 2drop r> drop exit then
     r> super to: ;m
     
   :m fill: -1 abort" fill: not allowed for sets" ;m
   :m insert: -1 abort" insert: not allowed for sets" ;m
;class 


[defined] floats [if]
:class fp-array <super 1-array
 ivar astart \ for shell sort
 fvar last

 :m init:  super init: last init:
     1 floats  elemSize ! \ set element size
     ['] f< sort-xt !
     ['] f= =xt !
     ;m

:m keep-duplicates: ( -- list ) \ returns a new sorted list instantiated in the heap
   self sort:
   self
   heap> fp-array 
   over each: drop last !: 0 locals| added list2 |
   begin
     dup each:
   while
     fdup last @: f= if added 0= if list2 add: -1 to added else fdrop then
               else last !: 0 to added
               then
   repeat drop list2 ;m 

:m remove-duplicates: ( -- list ) \ returns a new sorted list instantiated in the heap
   self sort:
   self
   heap> fp-array 
   over each: drop last !: locals| list2 |
   last @: list2 add:
   begin
     dup each:
   while
     fdup last @: f= if fdrop
                else fdup list2 add: last !: 
                then
   repeat drop list2 ;m 

 :m to: ( R: elem -- ) ( idx -- )  ^elem  f! ;m
 :m at: ( R: -- elem ) ( idx -- )  ^elem  f@ ;m
 :m .elem: ( R: elem -- ) f. ;m
 :m @elem: ( addr -- elem ) f@ ;m
 :m sum: ( F: -- r )
    0e
    begin
      self each:
    while
      f+
    repeat ;m
 :m average: ( F: -- r )
      self sum:
      #elems @ s>f f/ ;m

 :m max: ( F: -- max )
    self each: drop
    begin
     self each:
    while
    fover fover
     f< if fswap then fdrop
    repeat ;m
 :m min: ( F: -- min )
    self each: drop
    begin
     self each:
    while
    fover fover
     f> if fswap then fdrop
    repeat ;m

\ SHELLSORT     v1.3    28 October 1994         cgm     
\ (c) Copyright 1994 Charles G. Montgomery.  Permission is granted by the
\ author to use this software for any purpose provided this copyright
\ notice is preserved.
\ Shell-Metzger in-place ascending sort of floating array
\   see, e. g., Press et al, Numerical Recipes, Cambridge (1986)

: shellsort    ( nsize -- )

    0 self ^elem: astart !
    DUP
    BEGIN       ( nsize mspacing )
      2/ DUP    
    WHILE       ( n m )
      2DUP - 0
      DO        ( n m )
       0 I DO   ( n m )

\ compare Ith and (I+M)th elements 
              DUP I + DUP               ( n m l l )
              I FLOATS astart @ + F@      ( n m l ;F: Al Ai )
              FLOATS astart @ + F@        
              sort-xt @ execute                \ reverse this to get descending sort

\ switch them if necessary
              IF  DROP LEAVE            ( n m )
              THEN                      ( n m l )
              FLOATS astart @ + DUP F@
              I FLOATS astart @ + DUP F@  ( n m &Al &Ai ;F: Al Ai )
              SWAP F! F!

              DUP NEGATE                ( n m -m )
            +LOOP
      LOOP
    REPEAT   2DROP
;

 :m sort: #elems @  shellsort ;m
 :m p: \ pretty print
   ." f{ "
   self size: 0 ?do i self at: f. loop
   ." } " ;m

 :m indexOf: ( -- idx true | false ) ( F: r -- )
    self size: 0 ?do fdup i self at: f=
          if fdrop i true unloop exit then loop fdrop false ;m

;class
[then]

:class byte-array <super 1-array
 :m init:  super init:
     1 elemSize ! ;m
 :m to: ( elem idx -- )  ^elem  c! ;m
 :m at: ( idx -- elem )  ^elem  c@ ;m
 :m @elem: ( addr -- elem ) c@ ;m
;class

:class object-list <super 1-array
  :m delete: ( idx -- ) \ delete obj at idx and shrink list by 1 cell
     dup self at: <free
     ( idx ) super delete: ;m
  
  :m add: ( obj -- ) \ add any arbitrary object, it possibly will not have been cast in the heap (!!)
      size @ 0= if 0 super new: super uneach: then \ initialize ptr if necessary
     super size: dup >r 1+  super resize:
     r> super to: ;m

  :m xtadd: ( class-xt -- ) \ add any arbitrary object of given class, it will always be cast in the heap
     >body (heap) self add: ;m


fmscheck? [if]
  :m error?: self size: 0= abort" object-list is empty" ;m
  :m obj: ( -- ^obj ) \ returns last object added to list
     self error?:
     super size: 1- super at: ;m
  [else]
  :m obj: ( -- ^obj ) \ returns last object added to list
     super size: 1- super at: ;m
  [then]

  :m add!: ( ... class-xt -- ) \ add any arbitrary object of given class, it will always be cast in the heap
     self xtadd: self obj: !: ;m

  :m free: 
      ['] <free self map: \ objects in the list may be cast in the heap OR the dictionary
       self class_as> ptr free:
       self class_as> ptr init:
      ;m

  :m p:
        0 \ start counter at zero
      begin
        self each:
      while
        cr over . \ print counter
        p: \ print object
        1+ \ bump counter
      repeat drop ;m

 :m at@: ( -- <varies> )
    self at: @: ;m

 :m trim: ( n -- ) \ delete n items from end of list
    0 ?do self size: 1- self delete: loop ;m
 :m append: ( list-obj -- list-obj ) \ appends self to end of passed-in list
      begin
        self each:
      while
        over ( obj list-obj ) add:
      repeat ;m

   

;class


: upperCase? ( char -- flag ) \ flag is true if char is upper case
  [CHAR] A [ CHAR Z 1+ ] LITERAL WITHIN ;

\ Converts upper-case characters to lower case, modifying the contents
\ starting at addr for cnt chars
: to-lower   ( addr cnt -- )
 OVER \ addr cnt addr
 + SWAP  \ cnt+addr addr
 ?DO I C@ DUP
  upperCase?
  IF 32 + I C!
  ELSE DROP
  THEN
 LOOP ;

true value case-sensitive? \ used for searches in string objects
create bl$ bl c,

:class string+ <super string
  ivar start
  ivar end
  1 bytes theChar \ used for char searches
  1 bytes theChar2 \ used for char search and replace
  ivalue str1
  ivalue selfStrCopy
  ivar len
  ivar addr
  ivar current-idx
  
  
 :m init: super init: self uneach: ;m

 :m reset: 0  start !  0  end !  ;m
 :m !: ( addr len -- ) super !:  self reset: ;m
 :m start: ( -- pos-addr )  start ;m
 :m end: ( -- len-addr )  end ;m
 \ the number of chars remaining after the selection
 :m rem: ( -- remainder )  size @  end @ - ;m
 :m 1st: ( -- char ) \ return first char of substring 
     super ptr:  start @ + ( addr ) c@ ;m
 :m last: ( -- char ) \ return last char of substring 
    super ptr:  end @ + 1- ( addr ) c@ ;m
 
 :m copy: ( -- obj ) self @: heap> string+ dup >r !: r> ;m

  : ^char ( idx -- addr )
    dup 0< if size @ + then super ptr: + ;

fmsCheck? [if]

: ?idx: ( idx -- idx )
     dup
     abs size @ 
     < 0= ABORT" string+ index out of range" ; 
     

\ at: and to: indexes refer to the entire string
 :m at: ( idx -- char )
     ?idx:
     ^char c@ ;m
 :m to: ( char idx -- )
     ?idx:
     ^char c! ;m
[else]
 :m at: ( idx -- char )
     ^char c@ ;m
 :m to: ( char idx -- )
     ^char c! ;m
[then]

 :m uneach: 0  current-idx ! ;m

 :m each: ( -- char true | false)
     current-idx @ dup  size @ <
    if 1 current-idx +! self at: true else drop false self uneach: then ;m

:m upper: \ converts entire string to upper case
      self @: to-upper ;m

 :m lower: \ converts entire string to lower case
      self @: to-lower ;m

 :m @sub: ( -- addr len ) \ fetch substring
    super ptr:  start @ + ( addr )  end @  start @ - ;m

 :m selectAll: \ select entire string
    0  start !  size @  end ! ;m

\ compare addr1 len1 to the current *substring*
 :m =sub: ( addr1 len1 -- f ) \ observes case setting; true if equal, false if not
    DUP  end @  start @ - <> IF 2drop FALSE exit THEN  \ get out early if length mismatch
    len !  addr !
    case-sensitive?
    IF addr @ len @
       self @sub: COMPARE 0=
    ELSE
    heap> string+ heap> string+ to-iv str1  to-iv selfStrCopy
    self @sub: selfStrCopy !: \ make a temporary duplicate string object in the heap
    addr @ len @ str1 !: \ make a temporary duplicate string object in the heap
    selfStrCopy upper:   str1 upper: \ make both string copies uppercase
    selfStrCopy @:  str1 @: COMPARE 0=
    selfStrCopy <free  str1 <free
    THEN
    ;m
 
 :m =: ( obj -- true | false ) \ observes case setting; true if equal, false if not
     dup selectAll: @:
     self selectAll: self =sub:
     ;m
 
 : CIsearch: ( addr1 len1 -- true | false )
    heap> string+ heap> string+ to-iv str1 to-iv selfStrCopy len ! addr !
    self @: selfStrCopy !: \ make a temporary duplicate string object in the heap
     start @ selfStrCopy start: !    end @ selfStrCopy end: !
    addr @ len @ str1 !: \ make a temporary duplicate string object in the heap
    str1 @: len ! addr !
    selfStrCopy upper:   str1 upper: \ make both string copies uppercase
    
        selfStrCopy ptr: selfStrCopy end: @ + ( start-addr) selfStrCopy rem: ( len )
            addr @ len @ search  
        IF \ found
          ( u3 ) drop
          ( c-addr3 ) selfStrCopy ptr: - dup len @ +  end !  start !
          true
        ELSE  2drop false
        THEN
        str1 <free selfStrCopy <free \ dispose the temp strings
   ;

 \ Search: observes case setting and will hit on substrings.
 \ Subsequent searches will begin just after the prior substring hit.
 \ If the end of the string is reached then false will be returned
 \ even if the search for string has been hit and left alone.
 \ Must reset: to have subsequent search:'s start at beginning.
:m search: ( addr len -- true | false )
     >r 
     ( addr )
    \ first do rationality check
    \ return false if len to find is greater than remainder
      r@ self rem: > IF drop r> drop false exit THEN
    case-sensitive?
    IF  \ do case sensitive search
      ( addr) super ptr:  end @ + ( start-addr) self rem: ( len0)  rot ( addr) r@ ( len) search
      IF \ found
        ( u3 ) drop
        ( c-addr3 ) super ptr: - dup r> +  end !  start !
        true
      ELSE  2drop r> drop false
      THEN
    ELSE \ do case insensitive search
      r> ( addr len ) CIsearch:
    THEN
    ;m

:m chsearch: ( char -- flag ) \ observes case setting
    theChar c!   theChar 1 self search: ;m

 :m delete:  \ deletes the substring defined by START and END
             \ leaves START at the same place, END is set to START
      end @   start @ - 1 < IF exit THEN
     super ptr:  end @ +   \ src 
     super ptr:  start @ + \ src dest
     self rem: \ src dest cnt
     cmove 
     size @   end @  start @ - -  self resize:
     start @  end !
     ;m

\ inserts text starting at END, START and END are moved to end of inserted text
 :m insert: \ { addr len -- }
    >r \ addr on stack, len to return stack
     size @ r@ + self resize:
    super ptr:  end @ + dup ( src src ) r@ + ( dest ) self rem: r@ - ( cnt )
    move \ must use move, cannot use cmove 
    ( addr=src ) super ptr:  end @ + ( dest ) r@  cmove
     end @ r> + dup  start !  end ! ;m

\ inserts char at END, START and END are moved to just past inserted char
 :m chinsert: ( c -- )
     theChar c!   theChar 1 self insert: ;m 
 
\ replace selected text, if any, with given text
 :m replace: ( addr len -- )
     self delete: self insert: ;m

\ Search for text1 starting at END.
\ If found replace with text2.
\ Success flag is returned.
 :m sch&repl:  \ { addr1 len1 addr2 len2 -- flag } \ observes case setting
    >r >r
    ( addr1 len1 ) self search: dup
    IF r> r> ( addr2 len2 ) self replace:
    ELSE r> r> 2drop
    THEN ;m

\ Search for char1 starting at END.
\ If found replace with char2.
\ observes case setting
\ Success flag is returned.
 :m chsch&repl:  \ { char1 char2 -- flag }
    theChar2 c! theChar c!
    theChar 1 theChar2 1 self sch&repl: ;m

\ Reset self and replaces all occurrences of char1 with char2
\ in the WHOLE of self.  Self is lastly reset again.
 :m chreplall: \ { char1 char2 -- } \ observes case setting
    theChar2 c! theChar c!
	 self reset:
	 BEGIN
	   theChar 1 self search:
	 WHILE
	   theChar2 1 self replace:
	 REPEAT
	   self reset:
	;m    

\ Reset self and replaces all occurrences of (addr1 len1) by (addr2 len2)
\ in the WHOLE of self.  Self is lastly reset again.
 :m replall: \ { addr1 len1 addr2 len2 -- } \ observes case setting
	 self reset:
	 BEGIN
	   2over ( addr1 len1 ) self search:
	 WHILE
	   2dup  ( addr2 len2 ) self replace:
	 REPEAT 2drop 2drop
	 self reset:
 ;m    

\ remove all blanks
 :m remBLs:  bl$ 1 pad 0 self replall: ;m
 
 :m d:  \ formatted dump of string for debugging
   cr ."                     1                   2                   3                   4"
   cr ." 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0"
   cr  self size: 40 min 0 ?DO space i self at: emit LOOP
   cr  start @ 40 min 2* spaces [char] S emit
   cr  end @ 40 min 2* spaces [char] E emit
   cr ." start= "  start @ .  ."  end= "  end @ .  ."  rem= " self rem: . ."  size= "  size @ .
      ."   case-sensitive?= " case-sensitive? .
 ;m

:m skipChars: ( char -- ) \ observes case setting
    >r
    BEGIN
       end @  size @ = IF r> drop exit THEN
    
     end @ self at:
    case-sensitive? 0=
    IF dup lowerCase? IF 32 - THEN
    THEN
     r@ ( char ) =
    WHILE
      1  end +!
    REPEAT
      end @  start ! r> drop ;m

: (rdelimit): ( c -- flag ) \ true if a delimited word was found
    dup self skipChars:
    super ptr:  start @ + \ c addr
    self rem: \ c addr len
    over    \ c addr len addr
    +       \ c addr addr+len
    swap    \ c addr+len addr
    ?DO dup I C@ dup lowerCase? IF 32 - THEN
         = IF I super ptr: -  end ! UNLOOP drop TRUE exit THEN LOOP
    drop end @ self size: < if self size: end ! true
                            else false
                            then
    ; 

: CIrdelimit \ { c1 -- c2 } \ only called by rdelimit:
    dup ( c c ) lowerCase? IF  32 - THEN ;


\ find substring delimited by char on right and/or end of string
:m rdelimit: \ { c -- addr len true | false } \ observes case setting
    case-sensitive? 0=
    if ( c ) CIrdelimit then
     (rdelimit):
    if self @sub: true else false then
    ;m

 
\ find substring delimited by char and/or end of string
:m delimit: ( char -- addr len true | false )
    0 locals| strt c |
    c self chsearch:
    if end @ to strt
       c self chsearch:
       if start @ end !
          strt start !  
          self @sub: true exit
       else
          end @ start !
          self size: end !
          self @sub: true exit
       then
    then false ;m
 
\ find substring delimited by char1 and char2 and/or end of string
:m delimit2: ( char1 char2 -- addr len true | false )
    0 locals| strt c2 c1 |
    c1 self chsearch:
    if end @ to strt
       c2 self chsearch:
       if start @ end !
          strt start !  
          self @sub: true exit
       else
          end @ start !
          self size: end !
          self @sub: true exit
       then
    then false ;m

: post-trim
  self size: len @ - self resize: self reset: ;
  
:m trim-leading-nonalpha:
  0 len !
  self @: 2dup over + swap ?do i c@ isalpha if leave else 1 len +! then loop
  ( addr len ) >r dup len @ + ( src) swap ( src dest )
  r> len @ - cmove 
  post-trim ;m 

:m trim-trailing-nonalpha:
  0 len !
  self @: over + 1- ?do i c@ isalpha if leave else 1 len +! then -1 +loop
  post-trim ;m

:m trim:
   self trim-leading-nonalpha: self trim-trailing-nonalpha: ;m

: (capitalize) ( addr c flag n -- )
  swap if ( n ) + swap c! else 2drop drop then ;

: pre 1+ \ { beg end -- addr c flag }
  >r >r
  dup c@ dup r> r> within ;
 
:m capitalize:
   self ptr:  [char] a [char] z  pre -32 (capitalize) ;m

:m uncapitalize:
   self ptr:  [char] A [char] Z  pre  32 (capitalize) ;m

;class


:class file   
   ivar id  
   string name 
   ivar fam \ file access method 
   ivar open?
   string+ one-line \ used in read-line:
   object-list csv \ used in csv-read:

:m init: name init: one-line init: csv init: ;m
:m open?: ( -- flag ) open? @ ;m
:m name: ( -- ^obj ) name ;m
:m new:
   0 id !
   r/w fam !
   0 open? !
   ;m

:m open: ( caddr len -- ior )
   open? @ ABORT" open: attempt file already open"
   name !:
   name @: fam @ OPEN-FILE
   swap id !
   true open? !
   ;m

:m create: ( caddr len -- ior )
   2dup name !:
   fam @ CREATE-FILE
   swap id !
   true open? !
   ;m

\ Buffer-obj can be an object of *any* class that responds
\ appropriately to the new: and @: messages.  Here we use
\ a string object because it is convenient.
:m read: ( buffer-obj -- n ior ) \ n is actual bytes read
    >r
   open? @  0= ABORT" read: attempt file not open"
   id @ FILE-SIZE ABORT" read: FILE-SIZE failed"
   ( ud ) d>s r@ class_as> ptr new: \ note use of [class_as>]
   r> @: ( addr len ) id @ READ-FILE
   ;m

:m read-line: \ { m -- string+-obj true | false } \ m is max chars to be read per line
   open? @  0= ABORT" read-line: attempt file not open"
   one-line class_as> ptr new: one-line @: ( addr len )
   id @ READ-LINE throw
   if one-line resize:
      one-line reset: one-line true
   then ;m

:m csv-read: \ ( m -- object-list ) \ m is max chars per line (one line = one record) to be read
   begin
     dup self read-line:
   while \ line read succesful
      drop 
      heap> object-list >r \ one list for one line of data \ one line = one record
      begin
        [char] , one-line rdelimit:
      while
        \ string field successfully parsed
        heap> string+ r@ add: \ add a field object to the record
        ( one-line @sub: ) r@ obj: !: \ store text in last object
      repeat
        r> csv add: \ add one completed list (record) to the overall list of lists (records)
   repeat drop
    csv ;m

:m write-line: ( addr len -- ior ) 
   open? @  0= ABORT" write-line: attempt file not open"
   ( addr len ) id @ WRITE-LINE ;m

:m !pos: ( n -- )
   open? @  0= ABORT" !pos: attempt when file not open"
   s>d id @ REPOSITION-FILE ABORT" !pos: REPOSITION-FILE failed"
   ;m

:m @pos: ( -- n )
   open? @  0= ABORT" @pos: attempt when file not open"
   id @ FILE-POSITION ABORT" @pos: FILE-POSITION failed"
   d>s ;m

:m write: ( addr len -- ior )
   dup
   open? @  0= ABORT" write: attempt when file not open"
   ( len) s>d id @ RESIZE-FILE ABORT" write: RESIZE-FILE failed"
   id @ FLUSH-FILE drop
   0 self !pos: \ start write at beginning of file
   ( addr len) id @ WRITE-FILE
   ;m

:m close: ( -- ior )
   open? @  0= ABORT" close: attempt when file not open"
   id @ CLOSE-FILE
   0 open? !
   name free:  csv free: one-line free:
   ;m

\ file must be open before it can be closed
:m delete:  ( -- ior )
   name @: DELETE-FILE
   0 open? ! 
   name free:
   ;m
   
;class

:class file+ <super file
  string+ s
  
  :m init: s init: ;m

  :m read: ( addr len -- )
     self new:
     self open: throw
     s super read: 2drop
     self close: throw ;m

  :m write: ( addr len -- )
     self new:
     self create: throw
     s @: super write: throw
     self close: throw ;m
     
  :m @: ( -- obj ) s ;m
  :m free: s free: ;m

;class

0 [if]

file+ f
s" foo.txt" f read:
f @: upper:
s" bar.txt" f write: 
f free:

[then]

:class indexed    
  ivar curr-ind
  ivar each-idx
 :m .elem: ( elem -- ) . ;m
 :m at: ( idx -- elem )
    ^elem @ ;m
 :m to: ( elem idx -- )
    ^elem ! ;m
 :m init: 0 curr-ind !
     0 each-idx ! ;m
 :m add: ( elem -- )
    limit  curr-ind @ -  0= abort" can't add:, elements full"
    curr-ind @ self to:  1 curr-ind +! ;m
 :m uneach: 0 each-idx ! ;m
 :m each: ( -- elem true | false )
    each-idx @ dup curr-ind @ <
    if self at: true  1 each-idx +!
    else drop false
         each-idx @ curr-ind @ = if 0 each-idx ! then
    then ;m
 :m p: self size: 0 ?do cr i dup . self at: self .elem: loop ;m
 :m delete: ( idx -- ) \ delete elem at idx and shrink collection by 1
    curr-ind @ 0= abort" can't use delete: on an empty list"
    dup curr-ind @ <> \ no need to MOVE if we are deleting the last element
    if
     dup >r 1+ ^elem ( src)
           r@ ^elem ( dest) curr-ind @ r> - width * ( len) move
    else drop
    then
    -1 curr-ind +!
    ;m
 :m size: ( -- #elems ) curr-ind @ ;m
;class

:class ordered-col <super indexed  cell <indexed 
 :m search: ( val -- idx true | false )
    curr-ind @ 0 ?do dup i self at: = if drop i true unloop exit then loop drop ;m
;class


[defined] floats [if]

:class fordered-col <super indexed  1 floats <indexed
 :m .elem: ( F: elem -- ) f. ;m
 :m at: ( idx -- elem ) ^elem f@ ;m
 :m to: ( elem idx -- ) ^elem f! ;m
;class

[then]





:class 1-array2 <super 1-array
  :m p: \ pretty print
     ." i{ "
     self size: 0 ?do i self at: . loop
     ." } " ;m
;class

[undefined] parse-name [if]
: isspace? ( c -- f )
    bl 1+ u< ;

: isnotspace? ( c -- f )
    isspace? 0= ;

: xt-skip   ( addr1 n1 xt -- addr2 n2 ) \ gforth
    \ skip all characters satisfying xt ( c -- f )
    >r
    BEGIN
	dup
    WHILE
	over c@ r@ execute
    WHILE
	1 /string
    REPEAT  THEN
    r> drop ;

: parse-name ( "name" -- c-addr u )
    source >in @ /string
    ['] isspace? xt-skip over >r
    ['] isnotspace? xt-skip ( end-word restlen r: start-word )
    2dup 1 min + source drop - >in !
    drop r> tuck - ;
[then]

20 constant max-lstack-size
create stack max-lstack-size cells allot

variable lstack-depth
 
: >l ( n -- ) \ push item on stack
  lstack-depth @ max-lstack-size 1+ > ABORT" l-stack full"
  stack lstack-depth @ cells + !
  1 lstack-depth +! ;
: l@ ( -- n ) \ retrieve top stack item
  lstack-depth @ 0= ABORT" l-stack empty"
  stack lstack-depth @ 1- cells + @ ;
: l> ( -- n ) \ pop item from stack, last in first out
  l@  -1 lstack-depth +! ;

0 lstack-depth !

: cleanup-list ( list -- )
  <free
  lstack-depth @  0 ?do l> <free loop
  0 lstack-depth ! ; \ redundant?

: refilling-parse-name ( -- c-addr u )
	begin
	   parse-name dup 0= 
	while
	  2drop refill 0= -39 and throw
	repeat ;




: i{ ( -- 1-array ) 
  heap> 1-array2 locals| list |
	BEGIN
	   refilling-parse-name 2dup S" }" COMPARE  
	WHILE
	 ( addr len )
	 2dup s" .." compare 0=
	 if \ we have a range
	   2drop refilling-parse-name
	     2dup >integer
	     if
	       1+ ( end-of-range )
	       list last: 1+ ?do i list add: loop 2drop
	     else list <free true abort" invalid integer in i{...}"
	     then
     else
       >integer if ( n ) list add:
            else list <free true abort" invalid integer in i{...}"
            then
     then
	REPEAT ( addr len ) 2DROP list ; 



: $obj< \ { obj1 obj2 -- flag }
  swap @: rot @: compare 1 < ;


:class $1-array <super 1-array
 :m init:  super init: ['] $obj< sort-xt !
    ['] =: =xt !
 ;m
 :m p: ." ${ "
    begin
      self each:
    while
      p: space
    repeat ." } " ;m
  :m free:
     ['] <free self map: \ deallocate each string
     super free: ;m \ deallocate the ptr

\ from Hans Bezemer via comp.lang.forth June-26-2013
\ thread: Simple Sort - the smallest sorting routine?
\ ref:  http://de.wikipedia.org/wiki/Simplesort
: simplesort ( ^elem0 n -- )
  cells over + dup rot
  ?do
    dup i CELL+ ?do i @ j @ sort-xt @ execute
    if i @ j @ i ! j ! then 1 cells +loop
  1 cells +loop
  drop ;

  :m sort: 0 ^elem self size: simplesort ;m
    
 :m indexOf: ( addr len -- idx true | false )
    self size: 0 ?do 2dup i self at: @: compare 0=
          if 2drop i true unloop exit then loop 2drop false ;m

;class

: make-string ( addr len -- $obj )
  heap> string+ dup >r !: r> ;
: asList ( c obj -- list ) \ split the string using the given char, return a $1-array heap object
                            \ containing all of the split substrings
      dup reset:
      heap> $1-array >r
      2dup rdelimit: if make-string r@ add: else r> exit then
      begin
        2dup delimit:
      while
        \ string field successfully parsed
        make-string r@ add: 
      repeat 2drop
        r> ;

: ${ ( -- 1-array )  
  heap> $1-array
	BEGIN
	 ( obj )
	   refilling-parse-name 2dup S" }" COMPARE 
	WHILE
	 ( obj addr len )
      make-string over add:
	REPEAT ( obj addr len ) 2DROP ;


0 [if] optional FSL style array syntax, can be used with FORmula TRANslator 
\ probably biggest advantage is having dynamically resizeable arrays
\  and creating new arrays "on the fly" in the heap
: } ( ^obj idx -- ^elem )
  swap ^elem: ;
[then]

[defined] floats [if]  

:class fp-array2 <super fp-array
 :m p: \ pretty print
   ." f{ "
   self size: 0 ?do i self at: f. loop
   ." } " ;m

  :m >obj: ( -- list ) \ make each element of the array a fp object
                      \ and return a new list

  heap> object-list locals| list |
  begin
    self each:
  while
    ['] fvar list add!:
    list obj: !:
  repeat list ;m

;class

: f{ ( -- fp-array )
  heap> fp-array2
	BEGIN
	 ( obj )
	   refilling-parse-name 2dup S" }" COMPARE 
	WHILE
	 ( obj addr len )
	     >float if ( obj ) ( R: r ) dup add:
	            else ( obj ) <free true abort" invalid number in f{...}"
	            then
	REPEAT ( obj addr len ) 2DROP ;

[then]

0 [if] optional FSL style syntax
dict> fp-array value a{
3 a{ new:
1e a{ 0 } f!
2e a{ 1 } f!
3e a{ 2 } f!
a{ 0 } f@ f.
a{ 1 } f@ f.
a{ 2 } f@ f.
4 a{ resize:
4e a{ 3 } f!
a{ 3 } f@ f.
[then]

[defined] floats [if]


create '$ 1 c, char ' c,
:class string++ <super string+
 
 :m strip's:
   '$ count pad 0 self replall:
   ;m
 :m valid'?: ( -- flag ) \ true if string is bounded by "ticks"
    \ '' is not valid (a string of zero length), ' is not valid 
   self selectall:
   self 1st: [char] ' =
   self last: [char] ' = and ;m

 :m leading': ( -- flag )
   self selectall:
   self 1st: [char] ' =
   ;m

 :m p: [char] ' emit super p: [char] ' emit space ;m
 
;class



:class object-list2 <super object-list
 :m p: \ pretty print
    ." { " 
      begin
        self each:
      while
        p:
      repeat ." } " ;m
;class




: o{ ( -- list ) \ list allocated in the heap
\  heap> object-list2 0 locals| s list | \ Last Revision: 12 Jan 2017  08:14:13  dbh
  heap> object-list2 locals| list |
  begin
    refilling-parse-name
      begin
       2dup s" {" compare 0=
      while 
       2drop list >l recurse
      repeat
    2dup s" }" compare
      0= if lstack-depth @ 0>
            if list l> add:  2drop refilling-parse-name exit
            else false
            then
         else true
         then
  while
    ( addr len )
    2dup s" .." compare 0=
     if \ we have a range
       2drop refilling-parse-name
          2dup >integer
          if 
             1+ ( end-of-range )
             list last: 
             dup is-a var 0= abort" .. expected an integer range start"
             ( end-of-range var-obj )
             @: 1+ ?do i ['] var list add!: loop 2drop
          else
            2dup >float 
            if 
              1e f- f>s ( end-of-range )
             list last: 
             dup is-a fvar 0= abort" .. expected a fp range start"
             ( end-of-range fvar2-obj )
             @: 1+ f>s list last: @: ?do 1e f+ fdup ['] fvar list add!: loop fdrop
             2drop
            else
             2drop
            then 
          then
     else
	    2dup >integer
	      if  ['] var list add!: 2drop
	      else 2dup >float 
	             if  ['] fvar list add!: 2drop
	             else ['] string++ list add!:
	                  list last: dup >r valid'?: 0=
	                  r@ @: s" '" compare 0= or
	                     if
	                        r@ leading':
	                        if
	                          bl r@ +:
	                          [char] ' parse r@ add:
	                        else
	                          cr list last: p:  list cleanup-list
	                          r> drop true abort" invalid token in {..} list"
	                        then
	                     then
	                  r> strip's: 
	             then
	      then
	 then
  repeat 2drop
         lstack-depth @ if list cleanup-list
                           true abort" unmatched {..} pair in list"
                        then 
  list ;


:class hash-table-node
 ivar hash
 string key
 ivar val
 ivar next
 :m init: key init: ;m
 :m init ( val kaddr klen -- ) \ Initialise the node with the val, key
    key !:  val !  0 next ! ;m
 :m free: key free: ;m
 :m key@: ( -- obj ) \ will be a string object
    key ;m
 :m val@: ( -- val ) val @ ;m
 :m val!: ( -- val ) val ! ;m
 :m next: ( -- next-node ) next @ ;m
 :m next!: ( next-node -- ) next ! ;m
 :m @: ( -- hash ) hash @ ;m
 :m !: ( -- hash ) hash ! ;m
 :m d: \ dump
    cr ." hash " hash ?
    cr ." key " key p:
    cr ." val " val ?
    cr ." next " next ?
    cr ;m
;class

0 [if]
from https://en.wikipedia.org/wiki/Hash_table#In_programming_languages

Associative arrays
Hash tables are commonly used to implement many types of in-memory tables.  They are used to implement associative
arrays (arrays whose indices are arbitrary strings or other complicated objects), especially in interpreted
programming languages like Perl, Ruby, Python, and PHP. 

When storing a new item into a typical associative array and a hash collision occurs, but the actual keys themselves
are different, the associative array likewise stores both items.

\ The following is how class hash-table works: an insert: with an existing key will simply over-write the value
However, if the key of the new item exactly matches the key of an old item, the associative array typically erases the
old item and overwrites it with the new item, so every item in the table has a unique key. 

\ But see class hash-table-m for the ability to store multiple identical keys with differing values
\ and how to retrieve them later.

[then]


\ Hash collisions are handled by separate chaining with head records in the bucket array.
:class hash-table
 ivar table  \ will contain a 1-array object \ array of pointers to the nodes 
 ivar #nodes \ number of nodes
 ivar load \ load factor times 100
 ivar node
 ivar new-size
 ivar new-table
 ivar val
 ivar key-addr
 ivar key-len
 ivar idx
 ivar hash
 ivar last-node
 
 :m init: heap> 1-array table !
    0 #nodes !
    100 load !
    ( n ) table @ new: \ allocate table of n cells all initialized to zero
    0 table @ fill:
 ;m

\ idx to the table location is returned if false
 : do-search \ { -- node hash true | idx hash false }
    key-addr @ key-len @ hash-ivar dup hash ! table @ size: mod abs ( idx) dup idx ! table @ at:
    ( node-obj)
    begin
      dup if dup key@: @: key-addr @ key-len @ compare 0= if ( node-obj ) hash @ true exit then then
      dup
    while
      next:
    repeat drop idx @ hash @ false ;

: ll-delete \ { node -- true }
   dup >r
   key@: @: idx @ table @ at: key@: @: compare 0= if r@ next: idx @ table @ to:
                                                    r> <free  -1 #nodes +! true exit
                                                 then
   idx @ table @ at: \ walk linked-list, node to delete is not in table slot
   begin
    dup next: ?dup
   while
    ( parent-node child-node ) dup key@: @: r@ key@: @: compare 0=
    if \ found it, must relink the child of child-node to parent-node
       ( parent-node child-node ) next: swap next!: r> <free -1 #nodes +! true exit
    then
   repeat r> abort" failure in ll-delete" ; \ should never get here

:m delete: \ { kaddr klen -- flag } \ true if key was found (and hence deleted), false if not
   key-len ! key-addr !
   do-search
   if ( node hash ) drop ll-delete
   else ( idx hash ) 2drop false exit
   then ;m
 
 :m @: ( key-addr key-len -- val true | false )
    key-len ! key-addr !
    do-search
    if ( node hash ) drop dup last-node ! val@: true 
    else ( idx hash ) 2drop 0 last-node ! false
    then ;m

: transfer-insert \ { node -- }
    ( node) dup @: new-size @ mod abs ( idx ) new-table @ ^elem: \ compute new table location
    ( node-addr) dup @
    if  \ slot not empty, must add node-obj to hash collision linked-list 
      @ ( pre-existing node-obj in new table)
      begin  \ walk linked list to end
        dup next: ?dup
      while
        ( parent-node child-node ) nip 
      repeat 
        \ end of linked-list, add transferred node there
        ( node parent-node ) over swap  next!: \ link in node 
    else
      \ found empty slot in new-table
      over swap ! \ insert node in new-table
    then
    0 swap next!: \ zero out next in transferred node
    ;

: transfer \ { new-size new-table -- }
    new-table ! new-size !
    table @ size: 0 do i table @ at: \ step thru each element of the old table
    ?dup if ( parent-node) dup next: swap  transfer-insert  \ we have found a node in the old table
              \ note that we grab next before calling transfer-insert because transfer-insert
              \ will store 0 in the parent-node next, wiping out any existing child-node
              begin \ step thru linked list for additional nodes
               ?dup
              while
               dup next: swap  transfer-insert
              repeat
         then
    loop
    table @ <free 
    new-table @ table ! ;


\ new-node is a normal colon definition but is private to this class
: new-node ( -- new-node )
   1 #nodes +! \ increment #nodes
   heap> hash-table-node >r
   val @ key-addr @ key-len @ r@ init
   hash @ r@ !: r> ;

:m search: ( node hash -- true )
      drop \ drop hash
      val @ swap val!: true
   ;m

:m insert: \ { val key-addr key-len | idx node hash -- }
   key-len ! key-addr ! val !
   do-search
   if \ already present so just update val
      self search: if exit then
   then
   ( idx hash ) hash ! dup idx ! table @ at:
    ( node-obj) dup 0=
    if drop   \ found empty slot in table
       new-node  idx @ table @ to:
    else
    \ must add new node-obj to hash collision linked-list                 
      ( node-obj)  
      begin \ walk linked list to end 
        dup node !
        next: dup 0<>
      while
      repeat drop
       \ end of linked-list, add new node there
       new-node  node @ next!:
    then
       \ lastly test for need to increase table size
       \ and do it if required
       #nodes @ table @ size: load @ 100 */ >
       if table @ size: 2* 1+ ( new-size )
          heap> 1-array ( new-size new-table )
          2dup new:
          dup 0 swap fill:  
          transfer then
       ;m

 :m map: ( xt -- ) \ execute xt on each node
    begin
      table @ each:
    while
      ?dup if ( xt node ) 2dup ( xt node ) swap execute
              ( xt node )
              begin
                next: ( xt next ) ?dup
              while
                ( xt next ) 2dup swap execute
              repeat
           then
    repeat drop ;m

: (free) \ { node }
   >r
   r@ next: ?dup if ( next-node) recurse
                   then r> <free ;

:m free:
    begin
      table @ each:
    while
      ?dup if ( node ) (free)
            then
    repeat table @ <free ;m

: free-node-val \ { node }
  val@: <free ;

\ dump the entire hash table and each node
 :m d: table @ p:
     cr ." #nodes " #nodes ?
     cr ." table size " table @ size: .
     cr
    table @ size: 0 do
      cr i table @ at: dup 0=
      if .
      else dup d:
	      begin \ dump linked-list, if any linked nodes
	        dup
	      while
	        next: dup 0<> if dup d: then
	      repeat drop
	  then
    loop
    ;m

;class

 
0 [if]


.mem               
0 unFREEd pointers ok

hash-table t 
\ 1 t init 

8888 s" Now is the time." t insert:
45 s" witch" t insert:
88 s" goofy frog" t insert: 
99 s" dog" t insert: 
100 s" mama" t insert: 
101 s" boy-boy" t insert: 
509 s" 2nd amendment" t insert:
7 s" foo bar" t insert: 
601 s" Presidential" t insert:
 ok

t d:   
0 0 
1 12056656 
2 0 
3 0 
4 0 
5 0 
6 12417664 
7 0 
8 0 
9 11645648 
10 12416992 
11 0 
12 12389872 
13 0 
14 12021376 
#nodes 9 
table size 15 

0 

hash 112666 
key dog
val 99 
next 11725936 

hash -308396146 
key Now is the time.
val 8888 
next 12423904 

hash 1530006706 
key goofy frog
val 88 
next 0 

0 
0 
0 
0 

hash 1148838621 
key 2nd amendment
val 509 
next 0 

0 
0 

hash -2090872479 
key boy-boy
val 101 
next 0 


hash 4026460 
key mama
val 100 
next 0 

0 

hash -2016875292 
key Presidential
val 601 
next 0 

0 

hash -1232294759 
key foo bar
val 7 
next 12093184 

hash 145027679 
key witch
val 45 
next 0 
ok
.mem 
0 12423904 
1 12319568 
2 11725936 
3 11720448 
4 12093184 
5 11763616 
6 12416992 
7 12417056 
8 12056656 
9 12056720 
10 12389872 
11 12389936 
12 11645648 
13 11645712 
14 12417664 
15 12417728 
16 12021376 
17 12021440 
18 11666608 
19 12005904 
20 unFREEd pointers ok
t free: ok
.mem 
0 unFREEd pointers ok

[then]



\ Since the same join attribute, Name, occurs more than once
\ in both tables for this problem we need a hash table that
\ will accept and retrieve multiple identical keys if we want
\ an efficient solution for large tables. We make use
\ of the hash collision handling feature of class hash-table.

\ Subclass hash-table-m allows multiple entries with the same key.
\ After a @: hit one can inspect for additional entries with
\ the same key by using next: until false is returned.

:class hash-table-m <super hash-table

\ called within insert: method in superclass
 :m search: ( node hash -- idx hash false )
      swap drop idx @ swap false ;m
 :m next: ( -- val true | false )
    last-node @ dup
    if
      begin
       ( node ) next: dup
      while
        dup key@: @: key-addr @ key-len @ compare 0=
             if dup last-node ! val@: true exit then
      repeat 
    then ;m


;class
 
 
 
0 [if]
\ begin hash phase 
: obj ( addr len -- obj )
  heap> string+ dup >r !: r> ;

hash-table-m r   \ 1 r init 
s" Whales "   obj s" Jonah" r insert:
s" Spiders "  obj s" Jonah" r insert:
s" Ghosts "   obj s" Alan"  r insert:
s" Buffy "    obj s" Glory" r insert: 
s" Zombies "  obj s" Alan"  r insert:
s" Vampires " obj s" Jonah" r insert:
\ end hash phase

\ create Age Name table S
o{ { 27 'Jonah' }
   { 18 'Alan' }
   { 28 'Glory' }
   { 18 'Popeye' }
   { 28 'Alan' } } value s  

\ Q is a place to store the relation
object-list2 Q

\ join phase
: join \ { obj | list -- }
  0 locals| list obj |
  1 obj at: @: r @: \ hash the join-attribute and search table r
  if \ we have a match, so concatenate and save in q
    heap> object-list2 to list list q add: \ start a new sub-list in q
    0 obj at: copy: list add: \ place age from list s in q
    1 obj at: copy: list add: \ place join-attribute (name) from list s in q
    ( str-obj ) copy: list add: \ place first nemesis in q
    begin
      r next: \ check for more nemeses
    while
       ( str-obj ) copy: list add: \ place next nemesis in q
    repeat
  then ;

: probe
  begin
    s each: \ for each tuple object in s
  while
    ( obj ) join \ pass the object to function join
  repeat ;
 
probe \ execute the probe function

q p: \ print the saved relation

o{ 
{ 27 'Jonah' Whales Spiders Vampires } 
{ 18 'Alan' Ghosts Zombies } 
{ 28 'Glory' Buffy } 
{ 28 'Alan' Ghosts Zombies } } 


\ free allocated memory
s <free 
r free2:
q free:

.mem 
0 unFREEd pointers ok

[then]

0 [IF]
The basic tree engine is based on an example from Marcel Hendrix posted on usenet: 
 
  * DESCRIPTION : binary-tree 
  * AUTHOR      : Marcel Hendrix 
  * LAST CHANGE : July 29, 2006, Marcel Hendrix 

        REVISION -trees Binary trees        Version 1.00 

    From: 'Algorithms' Robert Sedgewick, ISBN 0-201-06672-6, page 178 - 185. 

Tree balancing has been added, in addition to making a tree class.  
[THEN]


:class node 
   ivar left
   ivar right
   ivar key
   ivar info
   
  :m left!: ( n -- ) left ! ;m
  :m right!: ( n -- ) right ! ;m
  :m key!: ( n -- ) key ! ;m
  :m info!: ( n -- ) info ! ;m
  :m left: ( -- n ) left @ ;m
  :m right: ( -- n ) right @ ;m
  :m key: ( -- n ) key @ ;m
  :m info: ( -- n ) info @ ;m
;class



:class btree 
  ivar size
  ivar oc-key \ container for keys ordered-col \ ordered-col will be sorted
  ivar oc-info \ container for info ordered-col, ordered as per key oc
  node head  
  node z 
  ivar key-display-xt
  ivar info-displ-xt
  ivar map-xt

:m init:
    0 size !
    0 head key!:
    0 head left!:
    z head right!:
    ;m

\ -- To search for a record with key v, do v 'head' tree-search ( -- link ) 
\ -- An unsuccessful search returns 'z'. 
:m search: \ { v | link -- link2 true | false }
  
  head >r
  dup ( v) z key!:
  BEGIN
    dup ( v) r@ key: <  IF   r> left: >r
          ELSE r> right: >r 
          THEN
    r@ key: over ( v key v) =
  UNTIL drop
   r> dup z = IF drop false ELSE true THEN
   ;m

\ -- To insert a node in the tree, we just do an unsuccessful search for it, 
\ -- then hook it on in place of 'z' at the point where the search was terminated. 
\ -- When a new node whose key is already in the tree is inserted, it goes 
\ -- to the right of the node already in the tree. All records with key equal to v 
\ -- can be processed by successively setting t to tree-search(v,t). 


: (insert) \ { v link | f -- link2 }
  0 locals| f link v |
    1 size +!
    BEGIN
     link to f 
     v link key: <  IF   link left: to link
           ELSE link right: to link 
           THEN
       link z =
    UNTIL
    heap> node to link
    v link key!:
    z link left!:
    z link right!:
    v f key: < IF   link f left!:
         ELSE link f right!:
         THEN
    link ;

:m insert: \ { key inf -- } \ inf=info, can be anything that fits in one cell
   swap ( inf key ) head  (insert) ( inf link2 ) info!: ;m

: (.nodes) \ { link -- }
         >r
         r@ z = IF r> drop EXIT THEN
         r@ key:  key-display-xt @ execute  r@ info:  info-displ-xt @ execute 
         r@ left: DUP z <> IF recurse ELSE DROP THEN 
         r> right: DUP z <> IF recurse ELSE DROP THEN
         ;

:m p: \ print the noces
     head right:  (.nodes) ;m

: (map) \ { link -- }
         >r
         r@ z = IF r> drop EXIT THEN
            r@ info:  \ get info to stack   
            map-xt @ execute  \ transform info 
            r@ info!:  \ place transformed stack item into info
         r@ left: DUP z <> IF recurse ELSE DROP THEN 
         r> right: DUP z <> IF recurse ELSE DROP THEN
         ;

:m map: ( xt -- ) \ apply xt to ( info @ ) for each node in the tree
     map-xt !
     head right:  (map) ;m

: (.sorted) \ { link -- }
         >r
         r@ z = IF r> drop EXIT THEN 
         r@ left:  recurse 
         r@ key:  key-display-xt @ execute 
         r> right:  recurse
         ;

:m sort: \ print sorted
    head right:  (.sorted)
         ;m

: (free) \ { link -- }
   >r
   r@ z = IF r> drop EXIT THEN
   r@ right: IF r@ right: recurse
          r@ left: recurse
          THEN
   r> <free ( cr ." free node" ) ;

: free-tree
    head right: (free)
    self init: ;
    
:m free:
    free-tree
   oc-key @ ?dup IF <free 0 oc-key ! THEN
   oc-info @ ?dup IF <free 0 oc-info ! THEN ;m

: (ord-col) \ { link -- }
         >r
         r@ z = IF r> drop EXIT THEN 
         r@ left:  recurse 
         r@ key: oc-key @ add:
         r@ info: oc-info @ add:
         r> right: recurse ;

: ord-cols ( -- oc-key oc-info ) \ ordered-cols will be sorted
   size @ heap> ordered-col oc-key !
   size @ heap> ordered-col oc-info !
   head right:  (ord-col) 
   oc-key @ oc-info @
   ;

: ((balance))
  oc-key @ size:  0 ?DO I oc-key @ at:  I oc-info @ at:  self insert:
              LOOP ;


: (balance) \ { | idx odd -- }
  0 0 locals| idx odd |
  oc-key @ size: 3 < IF ((balance)) exit THEN
  oc-key @ size: 2 /mod to idx to odd
  idx oc-key @ at: ( key) idx oc-info @ at: ( info) self insert:
  idx odd IF 1+ THEN 1
           DO idx I - oc-key @ at: ( key) idx I - oc-info @ at: ( info) self insert:
              idx I + oc-key @ at: ( key) idx I + oc-info @ at: ( info) self insert:
           LOOP
   odd 0=
   IF 0 oc-key @ at: ( key)  0 oc-info @ at: ( info)  self insert: THEN ;


:  balance
     ord-cols 2drop \ write out the tree to two ordered-col's (sorted)
     free-tree    \ clear the tree
    (balance) \ re-build the tree as a balanced tree
   oc-key @ <free oc-info @ <free \ free the ordered-cols
   0 oc-key !  0 oc-info !
   ; 

\ as a side-effect delete: will also balance the tree
:m delete: \ { n -- true | false } \ n = key to delete, true if successful
     ord-cols 2drop \ write out the tree to two ordered-col's (sorted)
   ( n ) oc-key @ search: ( idx T | F )
   IF ( idx ) dup oc-key @ delete:
                  oc-info @ delete:
        free-tree
        (balance)
      oc-key @ <free oc-info @ <free
      0 oc-key ! 0 oc-info !
      true
    ELSE
      false
    THEN ;m

\ a side effect of size: will be to also balance the tree
:m size: ( -- size ) size @
   balance ;m  

;class


0 [if]
cr .( Testing Binary Tree - With Recursion )

: display-char ." (" 0 .R ." ) " ; 


btree b 

' emit b iv key-display-xt ! 

' display-char b iv info-displ-xt ! 


cr .( insert t d e r a )
char t dup b insert: 
char d dup b insert:
char e dup b insert:
char r dup b insert:
char a dup b insert:

cr .( b p: )  b p: 
\ t(116) d(100) a(97) e(101) r(114) 

cr .( b size: . b p: ) b size: . b p: 
\ e(101) d(100) a(97) r(114) t(116) 

cr .( char e b delete: b p: )
T{ char e b delete: -> -1 }T
 b p: 
\ r(114) d(100) a(97) t(116) 

cr .( b sort: )  b sort: 
\ adrt

b free:


[undefined] noop [if] : noop ; [then]
' noop b iv key-display-xt !
' drop b iv info-displ-xt !

char t dup b insert:
char d dup b insert:
char e dup b insert:
char r dup b insert:
char a dup b insert:

T{ b p: -> 116 100 97 101 114 }T
T{ b size: . b p: -> 101 100 97 114 116 }T
T{ char e b delete: -> -1 }T
T{ b p: -> 114 100 97 116 }T
T{ b sort: -> 97 100 114 116 }T

b free:
[then]

cr .( entire FMS-SILib successfully loaded )
[then]