\ This software is free for use and modification by anyone for any purpose
\ with no restrictions or source identification of any kind.

\ Mar 2016 Douglas B. Hoffman
\ dhoffman888@gmail.com

0 [if]  

Based on the example program File Away! from Leo Brodie's book Starting Forth,
the paper version, not the (changed) online versions.

Required files to run this program:

FMS-SI.f     
FMS-SILib.f  
a-way-to-file.f (this file)
my-data-file.txt

User Words:

OPEN A data file must first be OPENed before it can be used.
  usage example: s" e:\FMS-SIFiles\my-data-file.txt" open

SAVE saves any changes and closes the file

ENTER Creates a new record and enters the provided data for the fields.
      Fields must be comma delimited and can be any length up to
      255 chars per field and 500 chars total for the record.
  usage: enter The Clown,Bozo,Circus Amusement Industry,(324)-555-1212 ext. 33

REMOVE Deletes the current record.

CHANGE Changes the current record specified field.
  usage: change job programmer
  
FIND Matches one input field/text against memory and shows name if found.
     Note that input is case insensitive and will hit on partial word matches.
     Also note that a space or spaces in the search for text, when enter or return
     is pressed, are significant. *important*
  usage: find job news=> Dan Rather    ( note partial word match )

GET Shows the text from the current record's specified field.
  usage: get job => newscaster

ANOTHER Will display next matching record's name beyond RECORD# if any.
  usage: another => Frank Reynolds

ALL Displays the entire database's matching field's names if any.
  usage: all => Dan Rather
                Frank Reynolds
                Jessica Savitch

PAIR Finds the record having a match on two field's contents, if any.
  usage: pair job newscast,phone 555=> Dan Rather

FULLNAME Matches to both surname and given name, if any.
  usage: fullname rath,dan=> Dan Rather   ( note partial word match and case insensitivity )



Programmer Notes: Application Structure

This program has been written to use any ANS Forth, normal files instead of
blocks files, and objects only in places where it makes sense.

Instead of a fixed maximum number of records and fixed maximum length of
fields, both are unlimited in size. Actually, the field sizes are limited
to length of 255 maximum, but this could easily be changed.  The records
are stored in a file with comma-delimited fields (the same format used to
ENTER new fields from the keyboard in the program).

Three kinds of objects are used and are listed as follows along with
the messages used with them in this program:

Object Class    Object Name       Messages Used
------------    -----------       -------------

file            f                  new:
                                   open:
                                   csv-read:
                                   write:
                                   close:

string+          s                 !:
                 what1             @:
                 what2             add:
                 (many nameless)   +:
                                   p:
                                   search:
                                   reset:
                                   free:

object-list      (many nameless)   size:
                                   add:
                                   obj:
                                   at:
                                   each:
                                   delete:


Message Use Descriptions
------------------------

file messages used
new: Prepares a file for use (initializes it).
open: ( caddr len -- ior ) Opens the file using the given file path\name.
csv-read: ( n -- obj ) Returns an object-list containing the entire
                       contents of the file parsed into comma separated
                       string objects. n is the maximum number of characters
                       per line that will be read.
write: ( obj -- ior )  Uses the passed in string object to overwrite the
                       entire file contents to disk.
close:  ( -- ior )     Closes the file.


string+ messages used
!: ( caddr len -- )    Stores the given text into the string object,
                       replacing all text previously in the string
                       if any.
@: ( -- caddr len )    Retrieves all text in the string object.
add: ( caddr len -- )  Adds the given text to the end of any text
                       already in the string, if any.
+: ( char -- )         Adds the given char to the end of any text
                       already in the string.
p:                     Prints the entire string to the display.
search: ( caddr len -- flag ) Observes case setting via the value
                       case-sensitive? and will hit on substrings.
                       Subsequent searches will begin just after the
                       prior substring hit.  If the end of the string
                       is reached then false will be returned.
reset:                 Resets the string search variables so that
                       new searches will start at the beginning of the
                       string.
free:                  When done with the string, frees any allocated
                       memory used by the string.


object-list messages used
size: ( -- n )         Returns the number of objects contained in the list.
add: ( obj -- )        Adds the given object to the end of the list by first 
                       increasing the size of the list by one.
obj: ( -- obj )        Returns the last object in the list.
at: ( index -- obj )   Returns the object corresponding to the zero-based
                       list index.
each: ( -- obj t | f ) Iterates through the entire list returning the next
                       object in the list, if any, and true.  If there are
                       no more objects then false is returned.
delete: ( index -- )   Deletes the specified, by index, object from the list
                       and reduces the size of the list by one.


If wanted, it would be simple to change the string searching behavior
to be identical to the original Starting Forth program.  That is, case sensitive
and whole-word matching. Just replace the search: method used in the
definition found? with =: and set the case-sensitive? value to true.
But case insensitive and partial word matches seem very convenient
for this database.  It would be trivial to provide a user word to
switch behavior whenever desired.

Note that the entire database resides either on disk or in an object-list
object stored in the global value rlist after we use OPEN.  From now on we
will refer to this list as rlist.  Rlist will contain any number of object-list
objects itself, each object-list corresponding to a database record, with each
object-list containing four string objects corresponding to the fields of the database.
We can visualize rlist as follows:

rlist = entire database after

   maxchars-per-record f csv-read: to rlist

has been executed in the word OPEN.


rlist is a list of objects:

obj(0)
obj(1)
obj(2)
...
obj(n)

Each of the above objects is an object-list itself and in this
program will be a list of four string objects as follows:

surname     given       job         phone
str-obj(0)  str-obj(1)  str-obj(2)  str-obj(3)

We access records and fields in the database with the at: message.

0 rlist at: ( -- obj(0) ) returns the first record in the database

We access a field in the given record by again using the at: message.

( obj(0) ) 0 swap at: ( -- string-obj(0) )  This will be the first field of the first record.
( string-obj(0) ) p: => Fillmore  \ we can display that field with the p: message.


So, restating the precise structure of the database after it has
been read from disk and resides entirely in RAM:

rlist = entire database = list of objects, each object is a list of 4 string objects


surname       given         job           phone

str-obj(0,0)  str-obj(0,1)  str-obj(0,2)  str-obj(0,3)
str-obj(1,0)  str-obj(1,1)  str-obj(1,2)  str-obj(1,3)
str-obj(2,0)  str-obj(2,1)  str-obj(2,2)  str-obj(2,3)
.
.
.
str-obj(n,0)  str-obj(n,1)  str-obj(n,2)  str-obj(n,3)

rlist size: => n+1 = number of records in database

To obtain a field from a record:
col row rlist at: at: @: ( caddr len ) \ see the word get-text


To store a new field in a record:
( caddr len ) col row rlist at: at: !: \ see the word change

To delete a record:
row rlist delete:  \ see the word remove  

To add a new record to the database:
heap> object-list rlist add:  \ see the word ENTER
\ the record will be added to the end of the list in rlist
To add a new field to the last record in the database:
kbd, heap> string+ dup >r !: r> rlist obj: add: \ see the word PUT



The field names surname, given, job, and phone are defined so
that executing them returns the column index ( 0, 1, 2, or 3 ) .

To maintain the syntax of the Starting Forth program we use
' >body @ to parse the column index from the field name in the
input stream.  See the word >field .



The rest of the program pretty much follows the flow of the
Starting Forth program.  Some of the definitions are identical
or nearly so.  One deviation is to use two text buffer string objects,
what1 and what2, and two corresponding variables, field1 and field2,
to "remember" the last one or two fields (field indexes) queried using words like
FIND and PAIR.  This is a consistent technique for "remembering"
both fields to search for.  The Starting Forth program uses a
very different-looking technique for remembering the second field.
Compare the definitions of PAIR and FULLNAME below to that from
Starting Forth. 

[then]

decimal


: field \ compile time: ( field# "field" -- field#+1 )
        \ child run time: ( -- field# )
  create dup , 1+ does> @ ;
0
field surname
field given
field job
field phone
1- constant max-field#

\ user configurable parameters
500 constant maxchars-per-record
10 constant cr#

\ named public objects
file f
string+ s
string+ what1 
string+ what2 

\ global variables
variable field1 \ Starting Forth used the name kind instead
variable field2 
variable record# 
0 value rlist
false to case-sensitive? 

: OPEN ( caddr len -- )  f new:  f open: throw
  maxchars-per-record f csv-read: to rlist ;

: >field ( "field" -- field# ) ' >body @ ;
: kbd, ( -- caddr len ) [char] , word count ;
: top 0 record# ! ;
: down record# @ 1+ rlist size: min record# ! ;
: get-text ( field# -- str-obj ) record# @ rlist size: 1- min rlist at: at: ;
: .field ( field# -- ) get-text p: space ;
: .name given .field surname .field ;
: keep1 ( field# -- ) field1 ! kbd, what1 !: ;
: keep2 ( field# -- ) field2 ! kbd, what2 !: ;
: found? ( caddr len field# -- flag ) get-text dup reset: search: ;
: -find ( -- flag ) rlist size: record# @ \ return true if not found, false if found
   ?do i record# !  what1 @: field1 @ found? if false unloop exit then
   loop true ;
: missing ." Not in file " ;
: (pair) rlist size: 0
  ?do i record# !
       -find if missing
             else what2 @: field2 @ found? if .name leave then
             then
  loop ;
: put kbd, heap> string+ dup >r !: r> rlist obj: add: ;
: ENTER ( "surname,given,job,phone" ) heap> object-list rlist add:
        put put put put ;
: REMOVE record# @ rlist delete: ;
: CHANGE ( "field" "text" -- ) >field get-text kbd, rot ( caddr len str-obj ) !: ; 
: FIND ( "field" "text" -- ) >field keep1 top -find if missing else .name then ;
: GET ( "field" -- ) >field .field ; 
: ANOTHER down -find if ." No other " else .name then ;
: ALL top begin cr -find 0= while .name down repeat ;
: PAIR ( "field" "text,field2" "text2" -- ) >field keep1  >field keep2 (pair) ;   
: FULLNAME ( "surname,given" -- ) surname keep1 given keep2 (pair) ;
: records->s  0 0 s !: 
   BEGIN
     rlist each:
   WHILE
     ( object-list[i] )
     max-field# 0 do i over at: @: s add: [char] , s +: loop \ add "surname,given,job,"
     max-field# swap at: @: s add:  cr# s +: \ add "phoneCR"
   REPEAT ;

: SAVE  records->s
   s @: f write: throw  f close: throw
   s free: what1 free:  what2 free:  ;
 

0 [if]
\ usage example:

s" /Users/doughoffman/Desktop/FMS-SIFiles/my-data-file.txt" open 


find job news
\ =>Dan Rather


all   
\ Dan Rather 
\ Jessica Savitch 
\ Frank Reynolds 

save

[then]